%start program
%token FUNC LONG FLOAT VOID IDENT NUM AND OR EXTERN SYSCALL IF WHILE SHIFTL SHIFTR RETURN NEQUAL EQUAL NOT BYTE

%{
  #include <stdio.h>
  #include <stdlib.h>
  #include <string.h> 
  #include "typedef.h"
  #include "code.h"
  #include <malloc.h>



  tsymbol * copy_list(tsymbol * src);
  void printTab(tsymbol * src, int returns);


  void trace(char * str) {
    if (debug)
      printf("%s", str);
    fflush(stdout);
  }

  tsymbol * symalloc() {
    tsymbol * ptr;
    ptr = NULL;
    ptr = (tsymbol *) malloc(sizeof(tsymbol));
    return ptr;
  }

  tsymbol * addelement(tsymbol * src, char * element, int type, char * regname, int isextern,int returnvalue, tsymbol * parameters ) {
    tsymbol * start = src;
    tsymbol * ptr = NULL;
    tsymbol * hptr = NULL;

    if (debug)
      printf("adding element: %s... ", element);
    trace("\n");

    if (src==NULL) {
      ptr = symalloc();
      start = ptr;
    }
    else {
      hptr = src->next;
      ptr = src;
      while (hptr!=NULL) {
	ptr = hptr;
	hptr = hptr->next;
      }
      fflush(stdout);
      ptr->next = symalloc();
      ptr = ptr->next;
    }
    
    ptr->name = (char *) malloc(sizeof(strlen(element)+1));
    strcpy(ptr->name,element);
    ptr->type = type;
    ptr->regname = regname;
    ptr->isextern = isextern;
    ptr->returnvalue = returnvalue;
    ptr->parameters = parameters;
    ptr->next = NULL;
    return start;
  }

  tsymbol * search1(tsymbol * src, char * element) {
    tsymbol * ptr;


    if (src==NULL)
      return 0;

    if (debug) {
      printf("Searching: %s...\n", element);
      fflush(stdout);
    }

    ptr = src;

    while (ptr != NULL) {
      if (strcmp(ptr->name, element)==0)
	return ptr;
      ptr=ptr->next;
      }
    
    return NULL;
  }


  tsymbol * uniqueinsert(tsymbol * src, char * name, int type, tsymbol * regname, int isextern, int returnvalue, tsymbol * parameters) {
    tsymbol * val;

    if (debug)
      printf("uniqueinsert:%s...\n", name);

    if (search1(src, name)!=NULL) {
      printf("Element %s is already defined. Quitting out\n",name);
      exit(1);
    }

    if (regname!=NULL)
      val = addelement(src,name,type,regname->name, isextern, returnvalue, parameters);
    else 
      val = addelement(src,name,type,NULL, isextern, returnvalue, parameters);

    if (val==NULL) {
      printf("Uniqueinsert of %s returns null!", name);
    }
    return val;
  }

  void checkident(tsymbol * list, char * name) {
    tsymbol * ptr;

    ptr = search1(list,name);
    if (ptr==NULL) {
      printf("Error: %s is not declared\n",name);
      exit(2);
    }
    if (ptr->type==FUNCTION) {
      printf("Error: %s is a function\n",name);
    }
  }

  int isextern(tsymbol * list, char * name ) {
    tsymbol * ptr;

    ptr = search1(list,name);
    if (ptr==NULL) {
      printf("Error: %s is not declared\n",name);
      exit(2);
    }

    if (ptr->isextern==1) {
      return 1;
    }
    return 0;

  }

  void checkfunction(tsymbol * list, char * name) {
    tsymbol * ptr;

    ptr = search1(list,name);
    if (ptr==NULL) {
      printf("Error: %s is not declared\n",name);
      printf("known things:");
      printTab(list,1);
      exit(2);
    }

    

    if (ptr->type!=FUNCTION) {
      printf("Error: %s is not a function\n",name);
    }
  }


  tsymbol copy_elements(tsymbol * src, tsymbol * dst) {
    dst->name = malloc(sizeof(char) * (strlen(src->name)+1));
    strcpy (dst->name, src->name);

    if (debug)
      printf("copied : %s\n",src->name);

    dst->type = src->type;
    
    if (src->regname!=NULL) {
      dst->regname = malloc(sizeof(char) * (strlen(src->regname)+1));
      strcpy (dst->regname, src->regname);
    }
    else
      dst->regname=NULL;
    
    dst->isextern = src->isextern;
    dst->returnvalue = src->returnvalue;

    if (src->parameters!=NULL) { 
      dst->parameters = copy_list(src->parameters);

    }
    else 
      dst->parameters = NULL;

    // setting the next pointer is done by copy_list
    // omitted
    dst->next=NULL;
    
  }

  tsymbol * merge_lists(tsymbol * first, tsymbol * second) {
    tsymbol * ptr;
    ptr = first;

    while (ptr->next!=NULL) {
      ptr=ptr->next;
    }

    ptr->next =second;

    return first;

  }

  // Kopiere Liste mit allem Drum und dran
  tsymbol * copy_list(tsymbol * src) {

    tsymbol * srcptr;
    tsymbol * dstptr;
    tsymbol * hptr;
    tsymbol * hptr1;

    if (src==NULL)
      return src;

    if (debug) {
      printf("\n***Starting of copy...");
      printTab(src,1);
    }
    srcptr = src;
    hptr =  symalloc();    
    dstptr = hptr;
    copy_elements(src, hptr);
    if (debug)
      printf("--------->START\n");

    // Jedes einzelne Element kopieren
    while (srcptr->next!=NULL) {
      srcptr= srcptr->next;
      if (debug)
	printf("&&&inserting %s\t",srcptr->name);
      hptr->next = symalloc();
      hptr=hptr->next;
      copy_elements(srcptr, hptr);
      if (debug) {
	printf("\n-------->next round..");
	printTab(dstptr,1);
      }
    }
    hptr->next = NULL;
    if (debug)
      printf("***End of copy..\n");
    return dstptr;
  }

  void printTab(tsymbol * src, int returns) {
    tsymbol * ptr = src;

    if (returns)
      printf("Start of symboltable--------\n");
    fflush(stdout);
    while (ptr!=NULL) {
      switch (ptr->type) {
      case FUNCTION:


	printf("function\t");

	if (ptr->isextern)
	  printf("extern ");
	
	switch (ptr->returnvalue) {
	case LONGTYPE:
	  printf("long ");
	  break;
	case FLOATTYPE:
	  printf("float ");
	  break;
	case LONGPTR:
	  printf("long * ");
	  break;
	case FLOATPTR:
	  printf("float * ");
	  break;
	case VOIDTYPE:
	  printf("void ");
	  break;
	case REGANN:
	  printf("regann detected in function - > forbidden -> quitting out\n ");
	  exit(1);
	  break;
	default:
	  printf("Error: Type is bad:%d!",ptr->returnvalue);
	  exit(1);
	}

	printf("%s (",ptr->name);
	if (ptr->regname!=NULL)
	    printf("( = %s) ",ptr->regname);

	printTab(ptr->parameters, 0);
	printf(");");
	break;
      case LONGTYPE:
	printf("long ");
	printf("%s",ptr->name);
	break;
      case FLOATTYPE:
	printf("float ");
	printf("%s",ptr->name);
	break;
      case LONGPTR:
	printf("long * ");
	printf("%s",ptr->name);
	break;
      case FLOATPTR:
	printf("float * ");
	printf("%s",ptr->name);
	break;
      case REGANN:
	printf("regann ");
	printf("%s",ptr->name);
       	break;
      default:
	printf("Error: unknown type, quitting out");
	exit(1);
      }

      if (ptr->regname!=NULL)
	printf("( = %s) ",ptr->regname);
      
      if (returns)
	printf("\n");
      else
	printf(", ");
      
      ptr=ptr->next;
    }

    if (returns)
      printf("End of symboltable--------\n");
  }
  
  treenode *newOp2(int op, treenode *left, treenode *right)
    {
      treenode *newNode = malloc(sizeof(treenode));
      
      if (newNode == NULL) yyerror("Malloc failed\n");
      
      newNode->op = op;
      newNode->left = left;
      newNode->right = right;
      
      return newNode;
    }
  
  treenode *newOp1(int op, treenode *left)
    {
      treenode *newNode = malloc(sizeof(treenode));
      
      if (newNode == NULL) yyerror("Malloc failed\n");
      
      newNode->op = op;
      newNode->left = left;
      newNode->right = NULL;
      
      return newNode;
    }
  
  treenode *newLeaf(int typ, long val, int reg, char * name, treenode * params, tsymbol * vardecls)
    {
      treenode *newNode = malloc(sizeof(treenode));
      
      if (newNode == NULL) yyerror("Malloc failed");
      
      newNode->op     = typ;
      newNode->left   = newNode->right = NULL;
      newNode->value  = val;
      newNode->regNr  = reg;
      newNode->name   = name;
      newNode->decls  = vardecls;
      if (debug)
      if (name!=NULL) {
	printf("\nAssigning %s to newNode->name",name);
	printf("\nNow newNode->name = %s, name = %s",newNode->name,name);
      }
      newNode->params = params;
      
      return newNode;
    }
  
  treenode *root;
  tsymbol * declarations;


%}

@attributes { long value; } NUM
@attributes { char * name; } IDENT
@attributes { tsymbol * ilist; tsymbol * slist;} globdecl globals
@attributes { treenode * ptr; } program
@attributes { tsymbol * ilist; tsymbol * slist;} pardecl pardecl1 regannotation vardecl 
@attributes { tsymbol * ilist; tsymbol * slist; treenode * ptr;} functionobj
@attributes { tsymbol * ilist; treenode * ptr;} statements statement expr params parameters
 
@attributes { int ivalue; int svalue; } optextern retval
@attributes { int optint; } optint 
@traversal @lefttoright @preorder LRpre
  
%%

program: globals functionobj   @{
           @e globals.ilist : ;
           @globals.ilist@ = NULL;
	   @i @functionobj.ilist@ = @globals.slist@;   
	   @e program.ptr: functionobj.ptr;
	   root = @program.ptr@ = @functionobj.ptr@;
	   @LRpre declarations = @functionobj.slist@;
	   //@LRpre printTab(@functionobj.slist@,1);	   
         @}
         ; 

globals: /* nothing */ @{
           @i @globals.slist@ = @globals.ilist@;
         @}
         | '[' globdecl ']' @{
           @i @globdecl.ilist@ = @globals.ilist@;
	   @i @globals.slist@ = @globdecl.slist@;
	   @}


globdecl:   /* nothing */ @{
	   @i @globdecl.slist@ = @globdecl.ilist@;
	   @}
           | LONG IDENT ';' globdecl @{
             @i @globdecl.1.ilist@ =  uniqueinsert(@globdecl.0.ilist@, @IDENT.name@,LONGTYPE,NULL,1,ILLEGAL,NULL);
	     @i @globdecl.0.slist@ = @globdecl.1.slist@;
           @}
	 | FLOAT IDENT ';' globdecl @{
             @i @globdecl.1.ilist@ =  uniqueinsert(@globdecl.0.ilist@, @IDENT.name@,FLOATTYPE,NULL,1,ILLEGAL,NULL);
	     @i @globdecl.0.slist@ = @globdecl.1.slist@;
           @}
	 | LONG '*' IDENT ';' globdecl @{
             @i @globdecl.1.ilist@ =  uniqueinsert(@globdecl.0.ilist@, @IDENT.name@,LONGPTR,NULL,1,ILLEGAL,NULL);
	     @i @globdecl.0.slist@ = @globdecl.1.slist@;
           @}
	 | FLOAT '*' IDENT ';' globdecl @{
             @i @globdecl.1.ilist@ =  uniqueinsert(@globdecl.0.ilist@, @IDENT.name@,FLOATPTR,NULL,1,ILLEGAL,NULL);
	     @i @globdecl.0.slist@ = @globdecl.1.slist@;
           @}
	 ;


 
	   
functionobj:  @{
	   @i @functionobj.slist@ = @functionobj.ilist@;  
	   @i @functionobj.ptr@ = NULL;
	   @}
           |  optextern retval IDENT '(' pardecl ')'  '{' vardecl statements '}' functionobj @{
	     @i @pardecl.ilist@ = NULL;
	     @i @retval.ivalue@ =0;
	     @i @optextern.ivalue@=0;
	     @i @functionobj.1.ilist@ = uniqueinsert(@functionobj.0.ilist@,@IDENT.name@, FUNCTION, NULL, @optextern.svalue@, @retval.svalue@, @pardecl.slist@);
	     @i @functionobj.0.slist@ = @functionobj.1.slist@;

	     @i @vardecl.ilist@ = merge_lists(copy_list(@functionobj.0.slist@),copy_list(@pardecl.slist@));
	     @i @statements.ilist@ = @vardecl.slist@;
	     @i @functionobj.0.ptr@ = newOp2(NEXTFUNC, newLeaf(FUNC_T,0,0,@IDENT.name@,@statements.ptr@,@vardecl.slist@),@functionobj.1.ptr@);

           @}
	     ;

optextern:    @{
	     @i @optextern.svalue@ = 0;
	     @}
             | EXTERN @{
	     @i @optextern.svalue@ = 1;
	     @}

retval:   VOID @{
	       @i @retval.svalue@ = VOIDTYPE;
	       @}
        | LONG @{
	       @i @retval.svalue@ = LONGTYPE;
	       @}
        | FLOAT @{
	       @i @retval.svalue@ = FLOATTYPE;
	       @}
        | LONG '*' @{
	       @i @retval.svalue@ = LONGPTR;
	       @}
        | FLOAT '*' @{
	       @i @retval.svalue@ = FLOATPTR;
	       @} 
        ;

vardecl:   @{
	  @i @vardecl.slist@ = @vardecl.ilist@;
	  @LRpre if (debug) printf("=======================\nin vardecl:");
	  @LRpre if (debug) printTab(@vardecl.ilist@,1);
	  @LRpre if (debug) printf("=======================\nend vardecl.");
	  @}

         | LONG IDENT ';' vardecl @{	   
	 @i @vardecl.1.ilist@ = uniqueinsert(@vardecl.0.ilist@,@IDENT.name@,LONGTYPE, NULL, 0,ILLEGAL, NULL);
	   @i @vardecl.0.slist@ = @vardecl.1.slist@;
	   @}
	 | FLOAT IDENT ';' vardecl @{	   
	 @i @vardecl.1.ilist@ = uniqueinsert(@vardecl.0.ilist@,@IDENT.name@,FLOATTYPE, NULL, 0,ILLEGAL, NULL);
	   @i @vardecl.0.slist@ = @vardecl.1.slist@;
	   @}
	 | LONG '*' IDENT ';' vardecl @{	   
	 @i @vardecl.1.ilist@ = uniqueinsert(@vardecl.0.ilist@,@IDENT.name@,LONGPTR, NULL, 0,ILLEGAL, NULL);
	   @i @vardecl.0.slist@ = @vardecl.1.slist@;
	   @}
	 | FLOAT '*' IDENT ';' vardecl @{	   
	 @i @vardecl.1.ilist@ = uniqueinsert(@vardecl.0.ilist@,@IDENT.name@,FLOATPTR, NULL, 0,ILLEGAL, NULL);
	   @i @vardecl.0.slist@ = @vardecl.1.slist@;
	   @}
	 ;

pardecl:    @{
	   @i @pardecl.slist@ = @pardecl.ilist@;
	   @}
         | LONG  IDENT regannotation pardecl1 @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.ilist@ = uniqueinsert(@pardecl.ilist@,@IDENT.name@,LONGTYPE, @regannotation.slist@, 0,ILLEGAL, NULL);
	   @i @pardecl.slist@ = @pardecl1.slist@;
	   @}
	 | FLOAT IDENT regannotation pardecl1  @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.ilist@ = uniqueinsert(@pardecl.ilist@,@IDENT.name@,FLOATTYPE, @regannotation.slist@, 0,ILLEGAL, NULL);
	   @i @pardecl.slist@ = @pardecl1.slist@;
	   @}
	 | LONG '*' IDENT regannotation pardecl1 @{
	   @e regannotation.ilist: ;
	      @regannotation.ilist@ = NULL;
	   @i @pardecl1.ilist@ = uniqueinsert(@pardecl.ilist@,@IDENT.name@,LONGPTR, @regannotation.slist@, 0,ILLEGAL, NULL);
	   @i @pardecl.slist@ = @pardecl1.slist@;
	   @}
	 | FLOAT '*' IDENT regannotation pardecl1 @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.ilist@ = uniqueinsert(@pardecl.ilist@,@IDENT.name@,FLOATPTR, @regannotation.slist@, 0,ILLEGAL, NULL);
	   @i @pardecl.slist@ = @pardecl1.slist@;
	   @}
	 ;

pardecl1:   @{
	   @i @pardecl1.slist@ = @pardecl1.ilist@;
	   @}
         | ',' FLOAT IDENT regannotation pardecl1  @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.1.ilist@ = uniqueinsert(@pardecl1.0.ilist@,@IDENT.name@,FLOATTYPE, @regannotation.slist@,0,ILLEGAL,NULL);
	   @i @pardecl1.0.slist@ = @pardecl1.1.slist@;
	   @}
         | ',' LONG  IDENT regannotation pardecl1 @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.1.ilist@ = uniqueinsert(@pardecl1.0.ilist@,@IDENT.name@,LONGTYPE, @regannotation.slist@,0,ILLEGAL,NULL);
	   @i @pardecl1.0.slist@ = @pardecl1.1.slist@;
	   @}
         | ',' FLOAT '*' IDENT regannotation pardecl1 @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.1.ilist@ = uniqueinsert(@pardecl1.0.ilist@,@IDENT.name@,FLOATPTR, @regannotation.slist@,0,ILLEGAL,NULL);
	   @i @pardecl1.0.slist@ = @pardecl1.1.slist@;
	   @}
         | ',' LONG  '*' IDENT regannotation pardecl1 @{
	   @i @regannotation.ilist@ = NULL;
	   @i @pardecl1.1.ilist@ = uniqueinsert(@pardecl1.0.ilist@,@IDENT.name@,LONGPTR, @regannotation.slist@,0,ILLEGAL,NULL);
	   @i @pardecl1.0.slist@ = @pardecl1.1.slist@;
	   @}
	 ;

regannotation: @{
	   @i @regannotation.slist@ = @regannotation.ilist@;
	   @}
           | '(' '=' IDENT ')' @{
	     @i @regannotation.slist@ = uniqueinsert(@regannotation.ilist@, @IDENT.name@, REGANN, NULL, 0,ILLEGAL,NULL); 
	     @}
          ;

statements: @{
	     @i @statements.ptr@ = NULL;
	     @}
	   |  statement statements @{	   
	     @i @statement.ilist@ = @statements.0.ilist@;
	     @i @statements.1.ilist@ = @statements.0.ilist@;
	     @i @statements.0.ptr@ = newOp2(FLOW_T, @statement.ptr@, @statements.1.ptr@);
	     @LRpre if (@statement.ptr@ == NULL) printf("statements: statement.ptr == NULL!********************");
	     @}
	 ;

statement:   '*' IDENT '=' expr ';' @{
            @i @expr.ilist@ = @statement.ilist@;
	    @LRpre checkident(@statement.ilist@,@IDENT.name@);
	    @i @statement.ptr@ = newOp2(ASS_T, newLeaf(MEMVAR_T, 0,0,@IDENT.name@,NULL,NULL),@expr.ptr@);
           @}
	   | '*' IDENT '=' '(' BYTE ')' expr ';' @{
            @i @expr.ilist@ = @statement.ilist@;
	    @LRpre checkident(@statement.ilist@,@IDENT.name@);
	    @i @statement.ptr@ = newOp2(ASS_T, newLeaf(MEMVAR_BYTE_T, 0,0,@IDENT.name@,NULL,NULL),@expr.ptr@);
           @}

           |  '*' NUM '=' expr ';' @{
             @i @expr.ilist@ = @statement.ilist@;	  
	     @i @statement.ptr@ = newOp2(ASS_T, newLeaf(MEMCON_T, @NUM.value@,0,NULL,NULL,NULL),@expr.ptr@);

            @}
	   | IDENT '=' expr ';' @{
            @i @expr.ilist@ = @statement.ilist@;
	    @i @statement.ptr@ = newOp2(ASS_T, newLeaf(VAR_T,0,0,@IDENT.name@,NULL,NULL),@expr.ptr@);
	    @LRpre checkident(@statement.ilist@,@IDENT.name@);
	    @}
           | IDENT '(' params ')' ';' @{
	     @i @params.ilist@ = @statement.ilist@;
	    @LRpre checkfunction(@statement.ilist@,@IDENT.name@);   	     
	    @i @statement.ptr@ = newLeaf(CALL_T, 0,0,@IDENT.name@, @params.ptr@,@statement.ilist@ );
	    //	    @LRpre if (@IDENT.name@ == NULL) printf("A functioncall with NULL IDENT!");
           @}
           | SYSCALL optint ';' @{
	       @i @statement.ptr@ = newLeaf(SYSCALL_T, @optint.optint@,0,NULL,NULL,NULL);
	      @}
           | IF '(' expr ')' statement  @{
             @i @expr.ilist@ = @statement.0.ilist@;
             @i @statement.1.ilist@ = @statement.0.ilist@;
             @i @statement.0.ptr@ = newOp2(IF_T,@expr.ptr@,newOp2(FLOW_T, @statement.1.ptr@,NULL));
           @}
           | IF '(' expr ')' '{' statements '}'  @{
             @i @expr.ilist@ = @statement.ilist@;
             @i @statements.ilist@ = @statement.ilist@;
             @i @statement.0.ptr@ = newOp2(IF_T,@expr.ptr@,@statements.ptr@);
           @}

           | WHILE '(' expr ')' '{' statements '}' @{
             @i @expr.ilist@ = @statement.ilist@;
             @i @statements.ilist@ = @statement.ilist@;
             @i @statement.ptr@ = newOp2(WHILE_T,@expr.ptr@, @statements.ptr@);
           @}
           | RETURN '(' expr ')' ';'  @{
             @i @expr.ilist@ = @statement.ilist@;	  
             @i @statement.ptr@ = newOp1(RETURN_T,@expr.ptr@);
            @}
           ;

optint:   @{
            @i @optint.optint@ = -1; 
         @}
         | NUM @{
            @i @optint.optint@ = @NUM.value@;
         @}
	 ;

expr:    IDENT @{
	   @LRpre checkident(@expr.ilist@,@IDENT.name@); 
           @i @expr.ptr@ = newLeaf(VAR_T, 0,0,@IDENT.name@,NULL,NULL);
	   @}
        | '*' IDENT @{
	   @LRpre checkident(@expr.ilist@,@IDENT.name@); 
           @i @expr.ptr@ = newLeaf(MEMVAR_T, 0,0,@IDENT.name@,NULL,NULL);
	   @}
        | '*' '(' BYTE ')' IDENT @{
	   @LRpre checkident(@expr.ilist@,@IDENT.name@); 
           @i @expr.ptr@ = newLeaf(MEMVAR_BYTE_T, 0,0,@IDENT.name@,NULL,NULL);
	   @}
        | '&' IDENT @{
	   @LRpre checkident(@expr.ilist@,@IDENT.name@); 
           @i @expr.ptr@ = newLeaf(MEMADDR_T, 0,0,@IDENT.name@,NULL,NULL);
	   @}
       | NUM @{
           @i @expr.ptr@ = newLeaf(CONST_T, @NUM.value@,0,NULL,NULL,NULL);
           @}
       | IDENT '(' params ')' @{
	     @LRpre checkfunction(@expr.ilist@,@IDENT.name@);
	    @i @params.ilist@ = @expr.ilist@;
	    @i @expr.ptr@ = newLeaf(CALL_T, 1,0,@IDENT.name@,@params.ptr@,NULL);
           @}
       | '(' expr '+' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(ADD_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr '*' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(MULT_T,@expr.1.ptr@,@expr.2.ptr@);

	 @}
       | '(' expr '-' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(SUB_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr '/' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(DIV_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr OR expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(OR_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr AND expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(AND_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr EQUAL expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(EQUAL_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr NEQUAL expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(NEQUAL_T,@expr.1.ptr@,@expr.2.ptr@);
	 @}
       | '(' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = @expr.1.ptr@;
	 @}
       | NOT expr  @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp1(NOT_T,@expr.1.ptr@);

	 @}

       | '(' expr SHIFTL NUM ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(SLT_T,@expr.1.ptr@,newLeaf(CONST_T,@NUM.value@,0,NULL,NULL,NULL));
	 @}
       | '(' expr SHIFTR NUM ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(SRT_T,@expr.1.ptr@,newLeaf(CONST_T,@NUM.value@,0,NULL,NULL,NULL));

	 @}
       | '(' expr '<' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(SMALLER_T,@expr.1.ptr@,@expr.2.ptr@);

	 @}
       | '(' expr '>' expr ')' @{
	 @i @expr.1.ilist@ = @expr.0.ilist@;
	 @i @expr.2.ilist@ = @expr.0.ilist@;
         @i @expr.0.ptr@ = newOp2(GREATER_T,@expr.1.ptr@,@expr.2.ptr@);

	 @}
       ; 

params:  @{
	   @i @params.ptr@ = NULL;
	   @}
        | NUM parameters @{
	   @i @parameters.ilist@ = @params.ilist@;
	   @i @params.ptr@ = newLeaf(NUM_PAR, @NUM.value@,0,NULL,@parameters.ptr@,NULL);
	  @} 
        | IDENT parameters @{
	   @LRpre checkident(@params.ilist@,@IDENT.name@);  
	   @i @parameters.ilist@ = @params.ilist@;
	   @i @params.ptr@ = newLeaf(NUM_IDE, 0,0,@IDENT.name@,@parameters.ptr@,NULL);
	  @}
	;
        

parameters:  @{
	  @i @parameters.ptr@ = NULL;
	  @}
	| ',' NUM parameters @{
	  @i @parameters.1.ilist@ = @parameters.0.ilist@;
	  @i @parameters.0.ptr@ = newLeaf(NUM_PAR, @NUM.value@,0,NULL,@parameters.1.ptr@,NULL);
	  @} 
        | ',' IDENT parameters @{
          @LRpre checkident(@parameters.ilist@,@IDENT.name@);  
	  @i @parameters.1.ilist@ = @parameters.0.ilist@;
	  @i @parameters.0.ptr@ = newLeaf(NUM_IDE, 0,0,@IDENT.name@,@parameters.1.ptr@,NULL);
          @}
; 


	   
%%

int line = 1;
/*
int main(void) {
	int ergebnis;

	ergebnis = yyparse();
	if (ergebnis==0) fprintf(stderr,"OK.\n");
	if (ergebnis==2) fprintf(stderr,"Syntaxfehler.\n");
	if (ergebnis==1) fprintf(stderr,"Lexikalischer Fehler.\n");
  return ergebnis;
}
*/

void printGlobalVariables() {
    tsymbol * ptr = declarations;

    printf("\n;Global Variables");
    printf("\n\t\tALIGN 16");
    while (ptr!=NULL) {
      switch (ptr->type) {
      case LONGTYPE:
	printf("\n%s:\t\tDD 0\t;LONG",ptr->name);
	break;
      case FLOATTYPE:
	printf("\n%s:\t\tDD 0\t;FLOAT",ptr->name);
	break;
      case LONGPTR:
	printf("\n%s:\t\tDD 0\t;LONG *",ptr->name);
	break;
      case FLOATPTR:
	printf("\n%s:\t\tDD 0\tFLOAT *",ptr->name);
	break;
      }
      ptr=ptr->next;
      printf("\n\t\tALIGN 16");

    }
    printf("\n\n");
}

int yyerror(char *msg) {
  fprintf(stderr,"Fehler in line %d -: %s\n",line,msg);
  exit(2);
}





